#include <stdio.h>
#include "graph.h"
#include "bubba.h"
#include "anneal.h"
#include <signal.h>

extern void InitRand();

struct channel *Chan;

#ifdef SIGNAL_HANDLER
void sig_int(int sno) {
  exit(1);
}
#endif

main(argc,argv)
 int argc;
 char *argv[];
{
 int length, nets;
 int seconds = -1;
 FILE *fil = NULL;
 char *infile;

#if defined(i386_unknown_solaris2_5) || defined(sparc_sun_solaris2_4)
 setmemwrite();
#endif

 printf("Bubba in main, argc=%d\n", argc);
 if (argc > 1)
    printf("file=%s\n", argv[1]);
 else
    printf("no args\n");
 
 /* Read command line */
 if (argc > 1) {
   infile = argv[1];
   fil = fopen(infile, "r");
   printf("open succeeded\n");
 }
 if (!fil) {
   fil = stdin;
 }

#ifdef SIGNAL_HANDLER
 if(signal(SIGINT, sig_int) == SIG_ERR) {
    printf("Exiting on signal install\n");
    exit(-1);
 }
#endif
 /* Initialize the Random numbers */
  InitRand(); 

 /* Get the Channel dimensions */
  nets = c_getint(fil,"Number of Nets",0,100,10);
  length = c_getint(fil,"Channel Length",0,100,12);

 /* Initialize the Channel */
  Chan = c_makechan(length,nets);

 /* Read in pins */
  c_readpins(fil,Chan);

 /* Make the subnets */
  c_makesnets(Chan);

#ifdef DEBUG
  a_subnetprint(Chan);
#endif

 /* Make the Vertical Constraint Graph */
  c_makegraph(Chan);

 /* Initialize the Partitions */
  CurP = p_init(Chan);
  TempP = p_init(Chan);
  BestP = p_init(Chan);
  BestP->numvalid = BIGINT;

 /* Initialize the screen */
#ifdef CURSES
  initscr();
#endif

 /* init timer */
  /* siginit(seconds);*/

 /* Do the annealing */
  a_anneal(Chan,CurP);

  exit(0);
}




#if defined(i386_unknown_solaris2_5) || defined(sparc_sun_solaris2_4)
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/procfs.h>
#include <sys/mman.h>
#include <stropts.h>

#define maxpmap 512     /* static const int maxpmap = 512; */
/* Vic's Solaris-specific code for getting complete coredumps from mutatees. */
/* Note that the mutatee must call this function (before it crashes!) */
/* Ideally this would be in the dyninst[API_]RT library loaded into the
   mutatee's address space, and called during DYNINSTinit, however, even
   that doesn't help when the attach causes the crash! */

/* Set every page in this process to be writable.  We hope this will
   cause pages with instrumented code to be saved in core files. */
/*static*/
int setmemwrite()
{
    int pfd, numpmap, i;
    prmap_t pmap[maxpmap];

    char buf[32];
    sprintf(buf, "/proc/%05d", getpid());
    pfd = open(buf, O_RDONLY);
    if (0 > pfd) {
	 perror("open (in setmemwrite)");
	 fprintf(stderr, "Can't open /proc on myself\n");
	 exit(1);
    }
    if (0 > ioctl(pfd, PIOCNMAP, &numpmap)) {
	 perror("PIOCNMAP (in setmemwrite)");
	 exit(1);
    }
    if (numpmap + 1 > maxpmap) {
	 fprintf(stderr, "Too many memory mappings\n");
	 exit(1);
    }
    if (0 > ioctl(pfd, PIOCMAP, pmap)) {
	 perror("PIOCMAP (in setmemwrite)");
	 exit(1);
    }
    for (i = 0; i < numpmap; i++) {
	 prmap_t *p = &pmap[i];
	 /* Enable WRITE if this region does not have it already and
	    we won't get in trouble for setting it (i.e., it is not
	    SHARED). */
	 if (~p->pr_mflags & MA_WRITE
	     && ~p->pr_mflags & MA_SHARED)
	      if (0 > mprotect(p->pr_vaddr, p->pr_size,
			       PROT_WRITE
			       | PROT_READ
			       | (p->pr_mflags & MA_EXEC ? PROT_EXEC : 0))) {
		   perror("mprotect (in setmemwrite)");
		   fprintf(stderr, "mprotect (it %d) args: %#010x, %x, %x\n",
			   i,
			   p->pr_vaddr, p->pr_size, 
			   PROT_WRITE
			   | PROT_READ
			   | (p->pr_mflags & MA_EXEC ? PROT_EXEC : 0));
		   exit(1);
	      }
    }
    close(pfd);
    return 0;
}
#endif
