#define SPACING ""
#define DASHES  ""

#include <stdio.h>
#include <stdlib.h>
/* #include <curses.h> */
#include "graph.h"
#include "bubba.h"

struct graph *VCG;

void c_density();

struct channel *c_makechan(len,nets)
  int len,nets;
{
  struct channel *ret;

  /* Allocate space for channel */
  if ((ret = (struct channel *) malloc(sizeof(struct channel))) == 0) {
    printf("Could not allocate space for channel\n");
    return(CHAN_PTR_ERROR);
  }

  /* Initialize */
  ret->length = len;
  ret->nets = nets;
  ret->numsnets = 0;
  ret->longchain = 0;
  ret->maxdensity = 0;

  /* Allocate pins */
  if ((ret->toppins = (int *) malloc(ret->length * sizeof(int))) == 0) {
    printf("MAKECHAN: Could not allocate top pins\n");
    free(ret);
    return(CHAN_PTR_ERROR);
  }

  if ((ret->botpins = (int *) malloc(ret->length * sizeof(int))) == 0) {
    printf("MAKECHAN: Could not allocate bottom pins\n");
    free(ret->toppins);
    free(ret);
    return(CHAN_PTR_ERROR);
  }

  /* Allocate space for vtracks */
  if ((ret->vtracks = c_maketracks(ret->length)) == 0) {
    printf("MAKECHAN: Could not make vertical tracks\n");
    free(ret->toppins);
    free(ret->botpins);
    free(ret->htracks);
    free(ret);
    return(CHAN_PTR_ERROR);
  }
  
  return(ret);
}
 

struct track *c_maketracks(num)
  int num;
{
  struct track *ret;
  int i;

  if ((ret = (struct track *) malloc(num * sizeof(struct track))) == 0)
   {
     printf("MAKETRACK: Could not allocate track!\n");
     return(TRACK_PTR_ERROR);
   }

  for (i = 0; i < num; i++) {
    ret[i].numruns = 0;
    ret[i].numalloc = 0;
    ret[i].runs = NULL;
  }
  return(ret);
}
    

void
c_readpins(fil, chan)
  FILE *fil;
  struct channel *chan;
{
  int i, length, net;

  length = chan->length;

 /* Read in top pins */
  printf("Top row pins:");
  for (i = 0; i < length; i++) {
    fscanf(fil,"%d",&net);
    printf("%2d ",net);
    chan->toppins[i] = net;
  }
  putchar('\n');

  printf("Bot row pins:");
  for (i = 0; i < length; i++) {
    fscanf(fil,"%d",&net);
    printf("%2d ",net);
    chan->botpins[i] = net;
  }
  putchar('\n');
  return;
}


int
c_getint(fil,str,min,max,defalt)
  FILE *fil;
  char *str;
  int min,max,defalt;
{
  int val = defalt;

  printf("%s:",str);

  fscanf(fil,"%d",&val);
 
  printf("%d\n",val);

  return(val);
}


void
c_newsnet(chan,parent,left,right)
  struct channel *chan;
  int parent, left, right;
{
  struct subnet *ptr;
  int top, bot;

  ptr = &chan->snet[chan->numsnets];

  ptr->parent = parent;
  ptr->left = left;
  ptr->right = right;
  ptr->length = right - left + 1;

  top = (chan->toppins[left] == parent);
  bot = (chan->botpins[left] == parent);
  if (top) ptr->lpin = TOP;
  if (bot) ptr->lpin = BOTTOM;
  if (top && bot) ptr->lpin = BOTH;


  top = (chan->toppins[right] == parent);
  bot = (chan->botpins[right] == parent);
  if (top) ptr->rpin = TOP;
  if (bot) ptr->rpin = BOTTOM;
  if (top && bot) ptr->rpin = BOTH;

  chan->numsnets++;

  return;
}


void
c_makesnets(chan)
  struct channel *chan;
{
  int i,temp[200],toppin,botpin;

  for (i = 1; i<=chan->nets; i++) {
    temp[i] = -1;
  }

  for (i = 0; i<chan->length; i++) {
    toppin = chan->toppins[i];
    botpin = chan->botpins[i];

   /* If toppin is in a valid net */
    if (toppin > 0) {
     /* If there is a left terminal to connect to */
      if (temp[toppin] >= 0) {
       /* Make a subnet */
        c_newsnet(chan,toppin,temp[toppin],i);
      }
     /* else if there is no left term to connect to */
      else {
       /* and the toppin and botpin need to be connected */
        if (toppin == botpin) {
         /* connect em with a subnet */
          c_newsnet(chan,toppin,i,i);
	}
      }
     /* Mark this as an available left term */
      temp[toppin] = i;
    }

   /* botpin is similar */
    if ((botpin > 0) && (botpin != toppin)) {
      if (temp[botpin] >= 0) {
        c_newsnet(chan,botpin,temp[botpin],i);
      }
      temp[botpin] = i;
    }
  }

  chan->width = chan->numsnets + 2; 

  /* Allocate space for htracks */
  if ((chan->htracks = c_maketracks(chan->width)) == 0) {
    printf("MAKECHAN: Could not make horizontal tracks\n");
    exit(1);
  }

  return;
}


void
c_constrain(chan,col,sdex) 
  struct channel *chan;
  int col;
  int sdex;
{
  struct subnet *thisnet;
  int i;
  int parent;
  int flag = 0;

  parent = chan->botpins[col];

  for (i = 0; i < chan->numsnets; i++) {
    thisnet = &chan->snet[i];
    if (thisnet->parent == parent) {
      if ((thisnet->left == col) || (thisnet->right == col)) {
        g_addadj(VCG,sdex,i);
        flag = 1;
      }
    }
  }

  if (!flag) {
    printf("CONSTRAIN: Big Prob, nobody below\n");
    exit(1);
  }

  return;
}

void
c_makegraph(chan)
  struct channel *chan;
{
  int i;
  struct subnet *ptr;

  if ((VCG = g_init(chan->numsnets)) == GRAPH_PTR_ERROR) {
    printf("MAKEGRAPH: Bogus return from g_init\n");
    exit(1);
  }

  for(i = 0; i < chan->numsnets; i++) {
    ptr = &chan->snet[i];
    if (ptr->lpin == TOP) {
      if (chan->botpins[ptr->left] != 0) {
       /* Figure out who it is above */
        c_constrain(chan,ptr->left,i);
      }
    }
    if (ptr->rpin == TOP) {
      if (chan->botpins[ptr->right] != 0) {
       /* Figure out who it is above */
        c_constrain(chan,ptr->right,i);
      }
    }
  }

  if (g_anycycle(VCG)) { 
    printf("There is a cycle in the Vert Cons Graph\n");
    printf("The channel is unroutable\n");
    exit(1);
  }

  chan->longchain = g_longest(VCG);

  c_density(chan);

  chan->minwidth = MAX(chan->longchain,chan->maxdensity);

  return;
}


void
c_makeruns(chan,part)
  struct channel *chan;
  struct partition *part;
{
  int trackno, left, right, setdex, snetdex;
  struct subnet *ptr;

  if (part->MG == GRAPH_PTR_ERROR) {
    p_makeMG(part);
  }

  g_toposort(part->MG);

  /* For each set in partition */
  for (setdex = 0; setdex < part->numvalid; setdex++) {
    trackno = part->MG->toposort[setdex];
    for (snetdex = 0; snetdex < part->numinset[part->setmap[setdex]]; snetdex++) {
      ptr = &chan->snet[part->set[part->setmap[setdex]][snetdex]];
      left = ptr->left;
      right = ptr->right;

      if (left != right) {
       /* Make the horizontal track */
        c_addrun(&chan->htracks[trackno],left,right);
  
       /* Make left vtracks */
        if ((ptr->lpin == TOP) || (ptr->lpin == BOTH)) {
          c_addrun(&chan->vtracks[left],0,trackno);
        }
        if ((ptr->lpin == BOTTOM) || (ptr->lpin == BOTH)) {
          c_addrun(&chan->vtracks[left],trackno,chan->width-1);
        }

       /* Make right vtracks */
        if ((ptr->rpin == TOP) || (ptr->rpin == BOTH)) {
          c_addrun(&chan->vtracks[right],0,trackno);
        }
        if ((ptr->rpin == BOTTOM) || (ptr->rpin == BOTH)) {
          c_addrun(&chan->vtracks[right],trackno,chan->width-1);
        }
      }
      else {
       /* Net has no horizontal track */
        c_addrun(&chan->vtracks[left], 0, chan->width-1);
      }
    }
  }
  return;
}


int
c_addrun(track,start,stop)
  struct track *track;
  int start,stop;
{
  track->numruns++;
  track->numalloc++;

  if (track->runs == NULL) {
    if ((track->runs = (struct run *) malloc(sizeof(struct run))) == 0) {
      printf("ADDRUN: Couldn't malloc\n");
      return(0);
    }
  }
  
  else {
    if ((track->runs = (struct run *) realloc(track->runs,sizeof(struct run) * track->numruns)) == 0) {
      printf("ADDRUN: Could not realloc\n");
      return(0);
    }
  }

  track->runs[(track->numruns) - 1].start = start;
  track->runs[(track->numruns) - 1].stop = stop;

  return(1);
}
  

void
c_clearruns(chan)
  struct channel *chan;
{
  int i;
  struct track *ptr;

  for (i = 0; i<chan->length; i++) {
    ptr = &chan->vtracks[i];
    if (ptr->numalloc > 0) {
      free(ptr->runs);
      ptr->runs = NULL;
      ptr->numalloc = 0;
      ptr->numruns = 0;
    }
  }

  for (i = 0; i<chan->width; i++) {
    ptr = &chan->htracks[i];
    if (ptr->numalloc > 0) {
      free(ptr->runs);
      ptr->runs = NULL;
      ptr->numalloc = 0;
      ptr->numruns = 0;
    }
  }

  return;
}


void    
c_density(chan)
  struct channel *chan;
{
  int temp1[200],temp2[200],temp3[200];
  int i,j,toppin,botpin,left,maxdensity=0,max=0;

  for(i=0;i<chan->length;i++) {
    temp3[i] = 0;
  }

  for(i=1;i<=chan->nets;i++) {
    temp1[i] = -1;
    temp2[i] = 0;
  }

  for(i=0;i<chan->length;i++) {
    toppin = chan->toppins[i];
    botpin = chan->botpins[i];

    if (toppin > 0) {
      if ((left = temp1[toppin]) >= 0) {
        for (j = left; j <= i; j++) {
          temp3[j]++;
	}
        if (temp2[toppin] > 0) {
          temp3[left]--;
	}
        temp2[toppin]++;
      }
      temp1[toppin] = i;
    }

    if ((botpin > 0) && (toppin != botpin)) {
      if ((left = temp1[botpin]) >= 0) {
        for (j = left; j <= i; j++) {
          temp3[j]++;
	}
        if (temp2[botpin] > 0) {
          temp3[left]--;
	}
        temp2[botpin]++;
      }
      temp1[botpin] = i;
    }
  }

  for(i=0;i<chan->length;i++) {
    if (temp3[i] > maxdensity) {
      maxdensity = temp3[i];
      max = i;
    }
  }

  chan->maxdensity = maxdensity;

  return;
}


void
c_printtrack(track)
  struct track *track;
{
  int i;
  struct run *run;

  printf("Track: %d runs %d alloc\n",track->numruns,track->numalloc);
  for (i = 0; i < track->numruns; i++) {
    run = &track->runs[i];
    printf("   (%d) ... %d to %d\n",i,run->start, run->stop);
  }

  return;
}
