#include <stdio.h>
#include <stdlib.h>
#include "graph.h"

/* Code for handling directed acyclic graphs */

static int scratch[256];
static int marker;

void g_tsort();
void g_free();


struct graph *g_init(num)
  int num;
{
  struct graph *ptr;
  int i;
 
  if ((ptr = (struct graph *) malloc(sizeof(struct graph))) == 0) {
    printf("G_INIT: Could not allocate graph!\n");
    exit(1);
  }

  ptr->nodes = num;
  ptr->max = num;

  if ((ptr->adj = (char **) calloc(sizeof(char *),ptr->nodes)) == 0) {
    printf("G_INIT: Could not calloc adj mat!\n");
    exit(1);
  }

  for (i = 0; i < ptr->nodes; i++) {
    if ((ptr->adj[i] = (char *) calloc(sizeof(char), ptr->nodes)) == 0) {
      printf("G_INIT: Could not allocate row %d\n",i);
      exit(1);
    }
  }

  if ((ptr->toposort = (int *) malloc(ptr->nodes * sizeof(int))) == 0) {
    printf("Could not allocate toposort\n");
    exit(1);
  }

  return(ptr);
}

struct graph *g_recalloc(g,num)
  struct graph *g;
  int num;
{
  int ind,dex;
  char *ptr;

  if ((g == GRAPH_PTR_ERROR) || (g == ((struct graph *) 0))) {
    return(g_init(num));
  }

  if (num > g->max) {
    printf("@");
    g_free(g);
    return(g_init(num));
  }

  g->nodes = num;
  for (ind = 0; ind < num; ind++) {
    ptr = g->adj[ind];
    for (dex = 0; dex < num; dex++) {
      *ptr = 0;
      ptr++;
    }
  }

  return(g);
}


void
g_toposort(g)
struct graph *g;
{
  int i;

/*   printf("Topo Sort: "); */

  marker = 1;

  for (i = 0; i < g->nodes; i++) {
    g->toposort[i] = 0;
  }

  for (i = 0; i < g->nodes; i++) {
    g_tsort(g,i);
  }

/*    putchar('\n');  */

  return;
}
 

void
g_tsort(g,node)
  struct graph *g;
  int node;
{
  int i;

 /* If this one has a label, go on */
  if (g->toposort[node]) return;

 /* Check out its ancestors first */
  for (i = 0; i < g->nodes; i++) {
    if (g_isadj(g,i,node)) {
      g_tsort(g,i);
    }
  }

 /* label it */
  g->toposort[node] = marker;
/*  printf("g->toposort[%d] = %d,",node,marker);  */

  marker++;

  return;
}


int /* bool */
g_cycfromnode(g,node)
  struct graph *g;
{
  int i;

  for (i = 0; i < g->nodes; i++) {
    scratch[i] = 0;
  }

  return(g_cycle(g,node));
}


int /* bool */
g_anycycle(g)
  struct graph *g;
{
  int i;

  for (i = 0; i < g->nodes; i++) {
    scratch[i] = 0;
  }

  for (i = 0; i < g->nodes; i++) {
    if (g_cycle(g,i)) {
/*      printf("G_ANYCYCLE: Found cycle starting from %d\n",i); */
      return(TRUE);
    }
  }

  return(FALSE);
}


int /* bool */
g_cycle(g,node)
  struct graph *g;
  int node;
{
  int i;

 /* See if this node was already visited */
  if (scratch[node] > 0) {
   /* If it is in a dead path then stop */
    if (scratch[node] == 2) {
      return(FALSE);
    }
   /* If it is in its own path then cycle */
    if (scratch[node] == 1) {
      return(TRUE);
    }

    printf("G_CYCLE: Big problem!\n");
    exit(1);
  }

 /* Since not visited, Mark it */
  scratch[node] = 1;

 /* Check out its children */
  for (i = 0; i < g->nodes; i++) {
    if (g_isadj(g,node,i)) {
      if (g_cycle(g,i)) {
        return(TRUE);
      }
    }
  }

  scratch[node] = 2;
  return(FALSE);
}


void
g_print(g)
  struct graph *g;
{
  int i, j;
  
  printf("ADJACENCY MATRIX\n");
  for (i = 0; i < g->nodes; i++) {
    for (j = 0; j < g->nodes; j++) {
      printf("%d ",(int) g->adj[i][j]);
    }
    putchar('\n');
  }

  return;
}


void
g_free(g)
  struct graph *g;
{
  int i;
  int num;

  num = g->nodes;
  for (i = 0; i < num; i++) {
    free(g->adj[i]);
  }

  free(g->adj);

  free(g->toposort);

  free(g);

  return;
}


int
g_longest(g)
  struct graph *g;
{
  int i,maxlen = 0,len;

  for (i = 0; i<g->nodes; i++) {
    scratch[i] = 0;
  }

  for (i = 0; i<g->nodes; i++) {
    len = g_length(g,i);
    if (len > maxlen) {
      maxlen = len;
    }
  }

  return(maxlen);
}


int
g_length(g,node)
  struct graph *g;
  int node;
{
  int i,maxlen=0,len;

  if (scratch[node] > 0) 
    return(scratch[node]);

  for(i=0;i<g->nodes;i++) {
    if (g_isadj(g,node,i)) {
      len = g_length(g,i);
      if (len > maxlen) {
        maxlen = len;
      }
    }
  }

  scratch[node] = maxlen + 1;
  return(scratch[node]);
}


void
g_copy(dst,src)
  struct graph *dst,*src;
{
  int in,out;

  for (out=0; out<src->nodes; out++) {
    dst->toposort[out] = src->toposort[out];
    for (in=0; in<src->nodes; in++) {
      dst->adj[out][in] = src->adj[out][in];
    }
  }

  return;
}
