#include <stdio.h>
#include <stdlib.h>
#include "anneal.h"
#include "graph.h"
#include "bubba.h"

struct partition *CurP;
struct partition *BestP;
struct partition *TempP;


void
p_hook()
{
}


struct partition *p_new(els)
  int els;
{
  struct partition *ret;
  int i;

  if ((ret = (struct partition *) malloc(sizeof(struct partition))) == 0) {
    printf("P_NEW: Could not allocate partition\n");
    exit(1);
  }

  if ((ret->numinset = (int *) calloc(sizeof(int),els)) == 0) {
    printf("P_NEW: Could not allocate numinset\n");
    exit(1);
  }

  if ((ret->setmap = (int *) calloc(sizeof(int),els)) == 0) {
    printf("P_NEW: Could not allocate setmap\n");
    exit(1);
  }

  if ((ret->set = (int **) calloc(sizeof(int *),els)) == 0) {
    printf("P_NEW: Could not allocate set\n");
    exit(1);
  }
  
  for (i = 0; i < els; i++) {
    if ((ret->set[i] = (int *) calloc(sizeof(int),els)) == 0) {
      printf("P_NEW: Could not allocate set %d\n",i);
      exit(1);
    }
  }

  ret->numsets = els;
  ret->numvalid = 0;
  ret->cost = 5000000.0;
  ret->MG = GRAPH_PTR_ERROR;

  return(ret);
}


struct partition *p_init(chan)
  struct channel *chan;
{
  int i;
  struct partition *ret;

  ret = p_new(chan->numsnets);

  for(i=0; i < ret->numsets; i++) {
    ret->set[i][0] = i;
    ret->numinset[i] = 1;
    ret->setmap[i] = i;
  }

  ret->numvalid = ret->numsets;

  return(ret);
}


void
p_makeMG(part)
  struct partition *part;
{
  int num,odex,idex,imem,omem;
  register int *omptr,*imptr,*osptr,*isptr;

  num = part->numvalid;

  /* Make a node for each set in partition */
  part->MG = g_recalloc(part->MG,num);

  /* Fill in adjacencies */
  for (omptr=part->setmap,odex = 0; odex < num; odex++,omptr++) {
    for (imptr=part->setmap,idex = 0; idex < num; idex++,imptr++) {
      if (idex == odex) continue; 
      for (osptr=part->set[*omptr],omem = 0; omem < part->numinset[*omptr]; omem++,osptr++) {
        for (isptr=part->set[*imptr],imem = 0; imem < part->numinset[*imptr]; imem++,isptr++) {
          if (g_isadj(VCG,*osptr,*isptr)) {
            g_addadj(part->MG,odex,idex);
	  }
	}
      }
    }
  }
  return;
}


int /* bool */
p_overlap(sn0,sn1)
  struct subnet *sn0,*sn1;
{
  int d0,d1,d2;

  if (sn0->parent == sn1->parent) {
    return(FALSE);
  }

  d0 = sn0->length;
  d1 = sn1->length;
  d2 = ((MAX(sn1->right,sn0->right)) - (MIN(sn0->left,sn1->left)) + 1);


  if (d2 < (d0 + d1)) { 
    return(TRUE);
  }

  return(FALSE);
}


int /* bool */
p_hconst(chan,part,realset) 
  struct channel *chan;
  struct partition *part;
  int realset;
{
  struct subnet *snet0, *snet1;
  int i,j,num;

  if ((num = part->numinset[realset]) > 1) {
    for (i=0; i<num; i++) {
      snet0 = &chan->snet[part->set[realset][i]];
      for (j=0; j<num; j++) {
        if (j <= i) continue;
        snet1 = &chan->snet[part->set[realset][j]];
        if (p_overlap(snet0,snet1)) {
          return(TRUE);
	}
      }
    }
  }

  return(FALSE);
}


int
p_whichset(map,num,val)
  int *map,val;
{
  int dex,*ptr;

  for (dex=0,ptr=map; dex<num; dex++) {
    if (*ptr == val) return(dex);
    ptr++;
  }
  return(-1);
}


int /* bool */
p_isvalid(chan,part,pmv)
  struct channel *chan;
  struct partition *part;
  struct pmove *pmv;
{
  int set1,set2;

  /* Construct merged-graph */
  p_makeMG(part);  

  /* Make sure that there are no cycles in mg */
  if (pmv->type == MV3) {
    return(TRUE);
  }

  /* Check for cycles in MG */
  set1 = p_whichset(part->setmap,part->numvalid,pmv->setdex1);
  set2 = p_whichset(part->setmap,part->numvalid,pmv->setdex2);

  if ((set1 >= 0) && g_cycfromnode(part->MG,set1)) {
    return(FALSE);
  }
  if ((set2 >= 0) && g_cycfromnode(part->MG,set2)) {
    return(FALSE);
  }

  /* Make sure that there are no horizontal constraints in any track */
  if ((p_hconst(chan,part,pmv->setdex1) || 
       p_hconst(chan,part,pmv->setdex2))) {
    return(FALSE);
  }

  return(TRUE);
}


void
p_printpart(part)
  struct partition *part;
{
  int i,j,setdex,*ptr;

  printf("*** Partition ***\n");
  printf("  %d sets (%d are valid)\n",part->numsets,part->numvalid);
  printf("  cost=%g\n",part->cost);
  printf("  Sets:\n");
  for (i = 0; i < part->numvalid; i++) {
    setdex = part->setmap[i];
    ptr = part->set[setdex];
    printf("    %d (%d): ",i,setdex);
    for (j = 0; j < part->numinset[setdex]; j++) {
      printf("%d ",ptr[j]);
    }
    printf("\n");
  }
  printf("*** End O Partition ***\n");
  printf("Channel width: %d\n",part->numvalid);
  return;
}


void
p_redosetmap(part) 
  struct partition *part;
{
  int i, setdex=0;

  for (i = 0; i < part->numsets; i++) {
    if (part->numinset[i] > 0) {
      part->setmap[setdex] = i;
      setdex++;
    }
  }

  if (setdex != part->numvalid) {
    printf("SETDEX != numvalid\n");
    p_hook();
    exit(1);
  }
  return;
}


int
p_delmem(part,setdex,memdex)
 struct partition *part;
 int setdex,memdex;
{
 int ret,num,*ptr,dex;

  ret = part->set[setdex][memdex];
  num = (--(part->numinset[setdex]));

  ptr = &part->set[setdex][memdex];
  for (dex = memdex; dex < num; dex++) {
    *ptr = *(ptr+1);
    ptr++;
  }

  return(ret);
}


void
p_copy(dest,src)
 struct partition *dest, *src;
{
 int i,j;

  dest->numsets = src->numsets;
  dest->numvalid = src->numvalid;
  dest->cost = src->cost;
 
  for (i = 0; i < src->numsets; i++) {
    dest->numinset[i] = src->numinset[i];
    dest->setmap[i] = src->setmap[i];
    for (j = 0; j < src->numinset[i]; j++) {
      dest->set[i][j] = src->set[i][j];
    }
  }

  dest->MG = g_recalloc(dest->MG,src->numvalid);

  g_copy(dest->MG,src->MG);

  return;
}
   
